#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <conio.h>
#include <io.h>

using namespace std;

#define MAX_LINE_LEN 8192
#define MAX_PGN_LINE_LEN 4096
#define MAX_PGN_MOVE_LINES 4096
#define MAX_PGN_HEADER_LINES 20

char line[MAX_LINE_LEN];

typedef char linetype[MAX_PGN_LINE_LEN];

enum fasetype { NOTHING, INHEADERS, INMOVES };

struct linesType
{
	linetype lines[MAX_PGN_MOVE_LINES];
   int number;
};

struct pgnType
{
   linesType headers;
   linesType moves;
} pgn;

int lineno;
fasetype currentfase;

//look if a string starts with 'search'
bool beginsWith(char *str, char *search)
{
	return strncmp(str, search, strlen(search)) == 0;
}

bool endsWith(char *str, char *search)
{
   str += strlen(str) - strlen(search);

	return strncmp(str, search, strlen(search)) == 0;
}

//look if a string starts with 'search'
bool beginsiWith(char *str, char *search)
{
	return strnicmp(str, search, strlen(search)) == 0;
}

//look if a string starts with 'search' and deletes it it finds
bool beginsWithAndDelete(char* &str, char *search)
{
	int lengte = (int)strlen(search);
	bool result = strncmp(str, search, lengte) == 0;

	if (result)
	{
		str += lengte;
	}

	return result;
}

//deletes starting spaces
void ltrim(char *str)
{
	while(str[0] == ' ')
	{
		strcpy(str, str+1);
	}
}

//deletes trailing spaces, returns, etc
void rtrim(char *str)
{
	int lengte = (int)strlen(str);

	while (lengte != 0)
	{
		if (str[lengte-1] == '\n' ||
			str[lengte-1] == '\r' ||
			str[lengte-1] == ' ') 
		{
			lengte--;
		}
		else
		{
			break;
		}
	}

	str[lengte] = '\0';
	
}

void stripper(char *str)
{
   rtrim(str);
   ltrim(str);
}

//make a string a certain length (with puting extra spaces)
void makelen(char *str, int lengte)
{
	int curlengte = (int)strlen(str);

	while (curlengte < lengte)
	{
		str[curlengte] = ' ';
		curlengte++;
	}

	str[lengte] = '\0';
}

//look how many times character 'c' occures in the string 'str'
int getcount(char *str, char c)
{
	int number = 0;

	for (int i=0; i<(int)strlen(str); i++)
	{
		if (str[i] == c) number++;
	}

	return number;
}

//get a string part of a string seperated with 'c'
//the first part is 0, the 2nd part 1, etc
void gettoken(char *str, char c, int searchnumber, char *buffer)
{
	int number = 0;
	char *strbuffer = buffer;

	for (int i=0; i<(int)strlen(str); i++)
	{
		if (str[i] == c) 
		{
			number++; 
		}
		else if (number == searchnumber)
		{
			*strbuffer = str[i];
			strbuffer++;
		}
	}

	*strbuffer = '\0';
}

//print the score list
void printLines(linesType* l)
{
	for (int i=0; i<l->number; i++)
	{
      cout << l->lines[i] << endl;
	}
}

void key_exit()
{
   char c;

   cerr << endl;
   cerr << "Press 'c' to continue or 'e' to exit" << endl;
   cin >> c;

   if (c == 'e') 
   {
      exit(1);
   }
}

void resetpgn()
{
   pgn.headers.number = 0;
   pgn.moves.number = 0;
   currentfase = NOTHING;
}

void uncomment()
{
   char comments[100];
   int commentnumber = 0;

   int printed_number_of_lines = 0;

	for (int i=0; i<pgn.moves.number; i++)
	{
      char *str = (char*)&pgn.moves.lines[i];

      int printed_number_of_chars = 0;

      while (*str != '\0')
      {
         char c = *str;

         if (c == '{') comments[commentnumber++] = '}';
         if (c == '[') comments[commentnumber++] = ']';
         if (c == '(') comments[commentnumber++] = ')';

         if (commentnumber > 0)
         {
            if (c == comments[commentnumber-1]) 
            {
               commentnumber--; 
            }
         }
         else if (printed_number_of_chars != 0 || c != ' ')
         {
            pgn.moves.lines[printed_number_of_lines][printed_number_of_chars] = c;
            printed_number_of_chars++;
         }

         str++;
      }

      if (printed_number_of_chars != 0)
      {
         pgn.moves.lines[printed_number_of_lines][printed_number_of_chars] = '\0';
         printed_number_of_lines++;
      }
	}

   pgn.moves.number = printed_number_of_lines;
}

void printpgn()
{
   uncomment();

   printLines(&pgn.headers);
   cout << endl;
   printLines(&pgn.moves);
   cout << endl;
   resetpgn();
}

bool isTag(char *&str)
{
   bool rc = false;

   static bool is_ended = true;

   if (str[0] != '[') 
   {
      //the former line was not ended ... but it is now ended
      //this to overcome the following:
      //
      //[Variation "Sicilian: Najdorf, Modern Main Line 13.f5 Nc5 14.f6: Perenyi's
      //16.Rg1"]
      if (!is_ended && endsWith(str, "]"))
      {
         return true;
      }
      else
      {
         return false;
      }
   }

   //I found [[White "Chess Tiger 11.4"] in the database
   if (beginsiWith(str, "[[White")) str++;

   if (beginsiWith(str, "[White")) rc = true; else
   if (beginsiWith(str, "[Black")) rc = true; else
   if (beginsiWith(str, "[Event")) rc = true; else
   if (beginsiWith(str, "[Site")) rc = true; else
   if (beginsiWith(str, "[PlyCount")) rc = true; else
   if (beginsiWith(str, "[Result")) rc = true; else
   if (beginsiWith(str, "[Opening")) rc = true; else
   if (beginsiWith(str, "[Round")) rc = true; else
   if (beginsiWith(str, "[ECO")) rc = true; else
   if (beginsiWith(str, "[NIC")) rc = true; else
   if (beginsiWith(str, "[Date")) rc = true; else
   if (beginsiWith(str, "[Time")) rc = true; else
   if (beginsiWith(str, "[Variation")) rc = true; else
   if (beginsiWith(str, "[SubVariation")) rc = true; else
   if (beginsiWith(str, "[Annotator")) rc = true; else
   if (beginsiWith(str, "[ICCResult")) rc = true; else
   if (beginsiWith(str, "[UTC")) rc = true; else
   if (beginsiWith(str, "[Termination")) rc = true; else
   if (beginsiWith(str, "[Source")) rc = true; else
   if (beginsiWith(str, "[FEN")) rc = true; else
   if (beginsiWith(str, "[SetUp")) rc = true; else
   if (beginsiWith(str, "[Board")) rc = true; else
   if (beginsiWith(str, "[Stage")) rc = true; else
   if (beginsiWith(str, "[Section")) rc = true; else
   if (beginsiWith(str, "[%")) rc = false; else
   {
      //
      // we reset the headers in the following line
      // since it does not support clean pgn output
      // 
      bool bReset = false;
      if (beginsiWith(str, "[d]")) bReset = true;
      if (beginsiWith(str, "[snip]")) bReset = true;
      if (beginsiWith(str, "[OPTIONS")) bReset = true;

      if (bReset)
      {
         resetpgn();
      }
   }

   if (rc)
   {
      is_ended  = (endsWith(str, "]"));
   }

   return rc;
}

void lower(char *str)
{
   for (int i=0; i < (int)strlen(str); i++)
      if (str[i] >= 'A' && str[i] <= 'Z')
         str[i] = str[i] - 'A' + 'a'; 
}

int getmonth(char *str)
{
   if (beginsiWith(str, "jan")) return 1;
   if (beginsiWith(str, "feb")) return 2;
   if (beginsiWith(str, "maa")) return 3;
   if (beginsiWith(str, "mar")) return 3;
   if (beginsiWith(str, "mrt")) return 3;
   if (beginsiWith(str, "apr")) return 4;
   if (beginsiWith(str, "may")) return 5;
   if (beginsiWith(str, "mai")) return 5;
   if (beginsiWith(str, "mei")) return 5;
   if (beginsiWith(str, "jun")) return 6;
   if (beginsiWith(str, "jul")) return 7;
   if (beginsiWith(str, "aug")) return 8;
   if (beginsiWith(str, "sep")) return 9;
   if (beginsiWith(str, "okt")) return 10;
   if (beginsiWith(str, "oct")) return 10;
   if (beginsiWith(str, "nov")) return 11;
   if (beginsiWith(str, "dec")) return 12;

   return -1;  
}

char lastchar(char *str)
{
   int lengte = (int)strlen(str);
   if (lengte == 0) 
      return '\0'; 
   else 
      return str[lengte-1];
}

bool determine(char *format, int getal[], int getalcounter, int& year, int& month, int& day)
{
   while (1)
   {
      if (year != -2 && month != -2 && day != -2)
      {
         if (year >= 0 && year < 100)
            year += 1900 + 100 * (year <= 10);

         //the year is not right
         if (year > 1000 && year < 1200 || year > 2005 || month > 12 || day > 31)
         {
            year = -1;
            month = -1;
            day = -1;
         }

         return true;
      }

      if (getalcounter == 0)
      {
         year = -1;
         month = -1;
         day = -1;
         continue;
      }

      if (getalcounter == 3 && getal[0] == -1 && getal[1] == -1 && getal[2] == -1)
      {
         year = -1;
         month = -1;
         day = -1;
         continue;
      }

      if (getalcounter == 2 && getal[0] == -1 && getal[1] == -1)
      {
         if (year == -2) year = -1;
         if (month == -2) month = -1;
         if (day == -2) day = -1;
         continue;
      }

      if (getalcounter > 0 && year != -2 && month != -2)
      {
         day = getal[0];
         continue;
      }

      if (getalcounter > 0 && year != -2 && format[0] == 'Y' && format[1] == ' ')
      {
         month = getal[1];
         day = getal[0];
         continue;
      }

      if (getalcounter == 1 && year == -2)
      {
         year = getal[0];
         if (month == -2) month = -1;
         if (day == -2) day = -2;
         continue;
      }

      //the year is known ... and to numbers are unknown
      //and one number is obviously a day
      if (getalcounter == 2 && year != -2)
      {
         if (getal[0] > 12 && getal[0] <= 31) 
         {
            month = getal[1];
            day   = getal[0];
            continue;
         }

         if (getal[1] > 12 && getal[1] <= 31) 
         {
            month = getal[0];
            day   = getal[1];
            continue;
         }
      }

      //the month is not set and one number is obviously a year
      if (getalcounter == 2 && month != -2)
      {
         if (getal[0] > 12 && getal[0] <= 99) 
         {
            year  = getal[1];
            day   = getal[0];
            continue;
         }

         if (getal[1] > 12 && getal[1] <= 99) 
         {
            day   = getal[0];
            year  = getal[1];
            continue;
         }
      }

      //we have a time
      //then format is yy-mm-dd yy:mm
      if (year != -2 && strstr(format, ":"))
      {
         month = getal[0];
         day   = getal[1];
         continue;
      }

      //we have a time
      //then format is yy-mm-dd yy:mm
      if (strstr(format, ":"))
      {
         year  = getal[0];
         month = getal[1];
         day   = getal[2];
         continue;
      }

      //we have 2 unknown numbers
      //if the last known is a year or the first then 
      //the middle is the month
      if (getalcounter == 2) 
      {
         if  (format[0] == 'Y')
         {
            month = getal[0];
            day   = getal[1];
            continue;
         }

         if (lastchar(format) == 'Y')
         {
            day   = getal[0];
            month = getal[1];
            continue;
         }
      }

      //1987.04.
      if (getalcounter == 1 && format[1] == '.' && year != -2)
      {
         if (month == -2)
         {
            month = getal[0];
            continue;
         }
      }

      if (getalcounter == 3)
      {
         if (getal[2] > 31 && getal[2] <= 99)
         {
            year = getal[2];
            getalcounter--;
            format[4] = 'Y';
            continue;
         }

         if (getal[1] > 31 && getal[1] <= 99)
         {
            year = getal[1];
            getal[1] = getal[2];
            getalcounter--;
            format[2] = 'Y';
            continue;
         }

         if (getal[0] > 31 && getal[0] <= 99)
         {
            year = getal[0];
            getal[0] = getal[1];
            getal[1] = getal[2];
            getalcounter--;
            format[0] = 'Y';
            continue;
         }
      }

      //02/28/02
      if (getalcounter == 3 && format[1] == '/')
      {
         year = getal[2];
         getalcounter--;
         format[strlen(format)-1] = 'Y';
         continue;
      }

      //we don't know this line ...
      year = -1;
      month = -1;
      day = -1;
   }

   return false;
}

void fixdate(char* str)
{
   char* org = str;
   int lengte = (int)strlen(str);
   int tmplen, formatlen=0;
   char tmp[100], format[100];
   int getalcounter = 0;
   int tmpgetal;
   int getal[5] = {0, 0, 0, 0, 0};
   int year = -2, month = -2, day = -2;

   //look through the date
   while (*str)
   {
      if (*str >= '0' && *str <= '9')
      {
         tmplen = 0;
         while (*str >= '0' && *str <= '9') {
            tmp[tmplen++] = *str;
            str++;
         } 
         tmp[tmplen] = '\0';

         //add year
         tmpgetal = atoi(tmp);

         if (tmpgetal >= 1500 && tmpgetal <= 2100) {
            year = tmpgetal;
            format[formatlen++] = 'Y';
         }
         else {
            format[formatlen++] = '9';
            getal[getalcounter++] = tmpgetal;
         }
      }
      else if (*str == '?')
      {
         tmplen = 0;
         while (*str == '?') {
            tmp[tmplen++] = *str;
            str++;
         } 
         tmp[tmplen] = '\0';

         if (tmplen == 4)
         {
            year = -1;
            format[formatlen++] = 'Y';
         }
         else {
            format[formatlen++] = '9';
            getal[getalcounter++] = -1;
         }
      }
      else if ((*str >= 'a' && *str <= 'z') || (*str >= 'A' && *str <= 'Z')) {
         tmplen = 0;
         while ((*str >= 'a' && *str <= 'z') || (*str >= 'A' && *str <= 'Z')) {
            tmp[tmplen++] = *str;
            str++;
         } 
         tmp[tmplen] = '\0';

         tmpgetal = getmonth(tmp);
         if (tmpgetal != -1) {
            month = tmpgetal;
            format[formatlen++] = 'M';
         }
      }
      else {
         format[formatlen++] = *str;
         str++;
      }
   }

   //close the format string
   format[formatlen] = '\0';

   //determine the right values
   determine(format, getal, getalcounter, year, month, day);

   //set the buffer back
   str = org;

   //print the date
   if (year == -1) str += sprintf(str, "????"); else str += sprintf(str, "%04d", year);
   str += sprintf(str, ".");
   if (month == -1) str += sprintf(str, "??"); else str += sprintf(str, "%02d", month);
   str += sprintf(str, ".");
   if (day == -1) str += sprintf(str, "??"); else str += sprintf(str, "%02d", day);
}

bool gettagvalue(char *tag, char *value)
{
   char *pos1;
   char *pos2;
   int lengte;

   *value = '\0';

   //look for the first "
   pos1 = strstr(tag, "\"");
   if (pos1 != NULL)
   {
      //look for the 2nd "
      pos2= strstr(pos1+1, "\"");
      if (pos2 != NULL)
      {
         lengte = (int)(pos2-pos1-1);
         strncpy(value, pos1+1, lengte);
         value[lengte] = '\0';
         stripper(value);

         return true;
      }
   }

   return false;
}

void addheader(char *str)
{
   char date[100];

   if (beginsiWith(str, "[Date "))
   {
      if (gettagvalue(str, date))
      {
         fixdate(date);
         sprintf(str, "[Date \"%s\"]");
      }
   }

   if (beginsiWith(str, "[EventDate "))
   {
      if (gettagvalue(str, date))
      {
         fixdate(date);
         sprintf(str, "[EventDate \"%s\"]");
      }
   }

   strcpy(pgn.headers.lines[pgn.headers.number], str);
   pgn.headers.number++;
}

void parseheader(char *str)
{
   if (pgn.headers.number >= MAX_PGN_HEADER_LINES)
   {
      cerr << "Error: Maximum number of headers reached: " << MAX_PGN_HEADER_LINES << endl;
      cerr << "Line(" << lineno << ")" << endl;
      cerr << "Tried to add : [" << str << "]" << endl;
      cerr << "Game will be dropped" << endl;
      resetpgn();
      return;
   }

   int lengte = (int)strlen(str);

   if (lengte > MAX_PGN_LINE_LEN)
   {
      cerr << "Error: Moveline is too long. " << lineno << endl;
      cerr << "The maximum length is " << MAX_PGN_LINE_LEN;
      cerr << " and this line is: " << lengte << endl;
      cerr << "Line(" << lineno << ")" << endl;
      cerr << "Tried to add : [" << str << "]" << endl;
      resetpgn();
      return;
   }

   //we have a header field without a starting '['
   //so we concatenate the headers together
   if (str[0] != '[')
   {
      sprintf(pgn.headers.lines[pgn.headers.number-1], "%s %s", pgn.headers.lines[pgn.headers.number-1], str);
   }
   else
   {
      addheader(str);
   }
}

bool addmoves(char *str)
{
   if (pgn.moves.number >= MAX_PGN_MOVE_LINES)
   {
      cerr << "Error: Maximum number of move lines reached: " << MAX_PGN_MOVE_LINES << endl;
      cerr << "Line(" << lineno << ")" << endl;
      cerr << "Tried to add : [" << str << "]" << endl;
      cerr << "Game will be dropped" << endl;
      printLines(&pgn.moves);
      resetpgn();
      return false;
   }

   int lengte = (int)strlen(str);

   if (lengte == 0) 
   {
      //we don't know where the game ends ...
      //se we just reset the whole game and wait for other games
      if (pgn.moves.number != 0)
      {
         resetpgn();
         return false;
      }
   }

   if (lengte > MAX_PGN_LINE_LEN)
   {
      cerr << "Error: The maximum length is " << MAX_PGN_LINE_LEN;
      cerr << " and this line is: " << lengte << endl;
      cerr << "Line(" << lineno << ")" << endl;
      cerr << "Tried to add : [" << str << "]" << endl;
      cerr << "Game will be dropped" << endl;
      resetpgn();
      return false;
   }

   strcpy(pgn.moves.lines[pgn.moves.number], str);
   pgn.moves.number++;

   if (endsWith(str, "1-0") || endsWith(str, "0-1") || endsWith(str, "1/2-1/2"))
   {
      return true;
   }
   else
   {
      return false;
   }
}


void replace(char *s, char *search, char *replace)
{
	while (1)
	{
		char *pos = strstr(s, search);

		if (pos == NULL) break;

		int lengte_search  = (int)strlen(search);
	   int lengte_replace = (int)strlen(replace);

		//first fix the end
		if (lengte_search > lengte_replace)
		{
			strcpy(pos+lengte_replace, pos+lengte_search);
		}

		if (lengte_search < lengte_replace)
		{
         printf("this function can't replace into longer strings");
         exit(1);
		}

		//then fix then new string
		strncpy(pos, replace, lengte_replace);
	}
}

void procesLine(char *str)
{
   replace(str, "<BR>", "");
   replace(str, "<P>", "");
   replace(str, "<br>", "");
   replace(str, "<p>", "");
   replace(str, "&gt;", ">");
   replace(str, "&quot;", "\"");

   while (*str == '>' || *str == ' ') str++;

   //we have an header
   if (isTag(str))
   {
      if (pgn.moves.number != 0)
      {
         printpgn();
      }

      currentfase = INHEADERS;
      parseheader(str);
   }
   else if (currentfase == INMOVES || currentfase == INHEADERS)
   {
      currentfase = INMOVES;

      if (addmoves(str))
      {
         printpgn();
         currentfase = NOTHING;
      }
   }
}

void parseLines(char *strFileName)
{
	FILE *f = fopen(strFileName, "r");
	if (f == NULL)
	{
		cerr << "Error opening " << strFileName << endl;
		exit(1);
	}

   lineno = 0; 
   resetpgn();

	while (fgets(line, MAX_LINE_LEN, f) != NULL)
   {
      lineno++;

		rtrim(line);
		ltrim(line);

		char *str = (char*)&line;

      procesLine(str);
	}

   fclose(f);
}

_finddata_t file;
int first=1;
char mask[512];

static int get_filename(char *wildcard) 
{
   static long link;
   int ret;
 
   do {
     ret=(first)? 0<=(link=(long)_findfirst(wildcard,&file)):!_findnext(link,&file);
     first=0;
   }
   while (ret && ((_A_HIDDEN | _A_SYSTEM | _A_SUBDIR) & file.attrib));
   return ret; 
}

int main(int argc, char* argv[])
{
	if (argc < 2)
	{
      cerr << "GETPGN v1.2 by Michel Langeveld 2003" << endl;
      cerr << "Please report errors to michel.langeveld@xs4all.nl" << endl << endl;
		cerr << "Usage: getscores <ascii-file>" << endl << endl;
      cerr << "Example: getscores *.txt" << endl;
		exit(1);
	}


   while (get_filename(argv[1]))
   {
      cerr << "Parsing " << file.name << " ... " << endl;
      parseLines(file.name);
   }

	return 0;

}

